/*
 * ===============================================================
 * IndepSetBDD - Implementation class
 *
 * TODO: memory for branching pool can be fixed, since there is
 *       a maximum number of nodes we will add at a time
 * ===============================================================
 */

#include <iostream>
#include "indepset_bdd.hpp"
#include <algorithm>

using namespace std;



//
// Constructor given a DIMACS instance filename
//
// IndepSetBDD(char* filename, int _maxWidth)
// DDX10:
IndepSetBDD::IndepSetBDD(
		const int _placeID,
		const int _ddWidth,
		const string & _instanceName,
		DDX10_Callback * const _cb
		)
	: DDX10_Base(_placeID, _ddWidth, _instanceName, _cb),
	  inst( new IndepSetInst ),
	  bestLB(0),
	  isExact(false),
	  isLBUpdated(false)
{
	assert( _ddWidth == INF_WIDTH || _ddWidth > 0 );

	inst->read_DIMACS(_instanceName.c_str());
	in_state_counter = new int[inst->graph->n_vertices];


	// ============================================================
	// Generate initial relaxation to derive variable ordering
	// TODO: save initial UB
	// ============================================================

	cout << "[BDD - place = " << x10_placeID << "] Computing variable ordering..." << endl;

	// set min-in-state variable ordering
	var_ordering = new MinInState(inst);
//	maxWidth = ROOT_WIDTH;
//
//	// generate relaxation
//	active_vertices.resize(inst->graph->n_vertices);
//	for (int i = 0; i < inst->graph->n_vertices; ++i) {
//		active_vertices[i] = i;
//	}
//	int ub1 = generateRelaxation(0);
//
//	// clean possible branch node pool
//	for (vector<BranchNode*>::iterator st = branch_nodes.begin(); st != branch_nodes.end(); ++st) {
//		delete (*st);
//	}
//	branch_nodes.clear();
//
//
//	// set new variable ordering
//	var_ordering = new RootOrdering(inst);
//	root_ordering.resize(inst->graph->n_vertices);
//	for (size_t i = 0; i < vertex_in_layer.size(); ++i) {
//		root_ordering[vertex_in_layer[i]] = i;
//	}

	// reinitialize width
	maxWidth = _ddWidth;
}


//
// Generate MDD relaxation. Returns upper bound.
//
// TODO: prune BDD nodes according to best lower bound (using perhaps some estimate?)
//
int IndepSetBDD::generateRelaxation(int initial_lp) {

//	cout << "max width = " << maxWidth << endl;

//	cout << "[BDD - place = " << x10_placeID << "] Node to explore:" << endl;
//	cout << "\tLP = " << initial_lp << endl;
//	cout << "\tState = { ";
//	for (vector<int>::iterator it = active_vertices.begin(); it != active_vertices.end(); ++it) {
//		cout << *it << " ";
//	}
//	cout << "}" << endl << endl;


	// ---------------------------------------------------------------------
	// 1. Initialization
	// ---------------------------------------------------------------------

	// clean list of local branch nodes
	localBranchNodes.clear();

	// Setup variable width
	maxWidth = MAX(10, 1000.0 / active_vertices.size());

	// Initialize internal structures for vertex ordering
	vertex_in_layer.clear();
	if (var_ordering->order_type == MinState) {
		for (vector<int>::iterator v = active_vertices.begin(); v != active_vertices.end(); ++v) {
			in_state_counter[*v] = 1;
		}

	} else if ( var_ordering->order_type == RootOrder ) {
		ComparatorAuxIntVectorDescending comp(root_ordering);
		sort( active_vertices.begin(), active_vertices.end(), comp );

	} else if( var_ordering->order_type == LexOrder ) {
		sort( active_vertices.begin(), active_vertices.end() );

	} else {
		cout << "Order undefined" << endl;
		exit(0);
	}


	// ---------------------------------------------------------------------
	// 2. Relaxation
	// ---------------------------------------------------------------------

	// Initialize pool for root node

	// create root node state
	State root_state;
	root_state.resize( inst->graph->n_vertices, false );
	for (vector<int>::const_iterator it = active_vertices.begin(); it != active_vertices.end(); ++it) {
		root_state.set(*it, true);
	}

	// create initial BDD node
	Node* initial_node = new Node(root_state, initial_lp, true);
	node_pool.clear();
	node_pool[ &(root_state) ] = initial_node;
	root_node = initial_node;

	BDDNodePool::iterator node_it, existing_node_it;
	Node* node;

	// relaxation control variables
	int current_vertex;
	int layer = 0;
	const int num_active_vertices = active_vertices.size();


	while ( layer < num_active_vertices ) {

		// ==================================
		// 1. Vertex and BDD node selection
		// ==================================

		// select next vertex and update active vertex list
		if (var_ordering->order_type == MinState) {
			current_vertex = choose_next_vertex_min_size_next_layer();

		} else if ( var_ordering->order_type == RootOrder ) {
			current_vertex = active_vertices.back();
			active_vertices.pop_back();


		} else if (var_ordering->order_type == LexOrder) {
			current_vertex = active_vertices.back();
			active_vertices.pop_back();

		} else {
			exit(0);
		}

		vertex_in_layer.push_back( current_vertex );
		assert( current_vertex != -1 );

		// Take nodes from the pool that have the current vertex in their state

		nodes_layer.clear();
		node_it = node_pool.begin();

		while (node_it != node_pool.end())	{

			if (node_it->second->state[current_vertex]) {

				// First case: state contains current vertex, thus we need to branch on it

				// if var ordering is min-in-state, decrement counter of active vertex list
				if (var_ordering->order_type == MinState) {
					for (vector<int>::iterator v = active_vertices.begin(); v != active_vertices.end(); ++v) {
						if ((*node_it->first)[*v]) {
							in_state_counter[*v]--;
						}
					}
				}

				// move from the pool to current layer list
				nodes_layer.push_back(node_it->second);
				node_pool.erase(node_it++);

			} else {
				// Second case: state does not contain vertex; we can maintain it in the pool
				++node_it;
			}
		}

		// PRINT LAYER
//		cout << "Layer " << layer << " - current vertex: " << current_vertex;
//		cout << " - pool size: " << node_pool.size();
//		cout << " - before merge: " << nodes_layer.size();
//		cout << " - total: " << node_pool.size() + nodes_layer.size();
//		cout << endl;


		// ==================================
		// 2. Node merging
		// ==================================

		if( maxWidth != INF && (int)nodes_layer.size() > maxWidth ) {
			mergeLayer(layer, nodes_layer);
		}


		// ==================================
		// 3. Branching
		// ==================================

		Node* branch_node;
		for (vector<Node*>::iterator it = nodes_layer.begin(); it != nodes_layer.end(); ++it) {

			branch_node = (*it);

			// TODO: CHECK THIS!
			if (!branch_node->state[current_vertex]) {
				delete branch_node;
				continue;
			}


			branch_node->state.set(current_vertex, false);

			// --------------- One arc ---------------------

			node = new Node(branch_node->state,
					branch_node->longest_path + inst->graph->weights[current_vertex],
					branch_node->exact);

			// remove adjacency vertices
			node->state &= inst->adj_mask_compl[current_vertex];

			// estimate size of the independent set from this vertex. We create it only if it
			// can improve current best upper bound
			int estimate_weight = branch_node->longest_path + inst->graph->weights[current_vertex];
//			for (size_t val = node->state.find_first(); val != state_end; val = node->state.find_next(val)) {
//				estimate_weight += inst->graph->weights[val];
//			}
			estimate_weight += node->state.count();

			if (estimate_weight > bestLB) {

				// Equivalence test: check if node is in list
				existing_node_it = node_pool.find( &(node->state) );
				if (existing_node_it != node_pool.end()) {
					// node already exists in the pool: update node match
					merge(node, existing_node_it->second);
					node = existing_node_it->second;

				} else {
					node_pool[ &(node->state)] = node;

					// update eligibility list
					if (var_ordering->order_type == MinState) {
						for (vector<int>::iterator v = active_vertices.begin(); v != active_vertices.end(); ++v) {
							if ((node->state)[*v]) {
								in_state_counter[*v]++;
							}
						}
					}
				}
			} else {
				delete node;
			}


			// --------------- Zero arc ---------------------

			// we can use the branch node for the zero arc, since its state was already updated
			node = branch_node;

			// We create a zero arc only if it can improve best lower bound
			// estimate size of the independent set from this vertex. We create it only if it
			// can improve current best upper bound
			estimate_weight = branch_node->longest_path;
//			for (size_t val = node->state.find_first(); val != state_end; val = node->state.find_next(val)) {
//				estimate_weight += inst->graph->weights[val];
//			}
			estimate_weight += node->state.count();

			if (estimate_weight > bestLB) {
				existing_node_it = node_pool.find( &(node->state) );
				if( existing_node_it != node_pool.end() ) {
					// node already exists in the pool: update node match
					merge(node, existing_node_it->second);
					node = existing_node_it->second;

				} else {
					node_pool[ &(node->state)] = node;

					// update eligibility list
					if (var_ordering->order_type == MinState) {
						for (vector<int>::iterator v = active_vertices.begin(); v != active_vertices.end(); ++v) {
							if ((node->state)[*v]) {
								in_state_counter[*v]++;
							}
						}
					}
				}
			} else {
				delete node;
			}
		}

		// if the number of nodes in the pool is empty, then we do not need to explore this BDD further
		// since there are no better feasible solutions that can be generated from here
		if (node_pool.empty()) {

			for(vector<BranchNode*>::iterator it = localBranchNodes.begin(); it != localBranchNodes.end(); ++it) {
				delete (*it);
			}
			localBranchNodes.clear();

			// reset internal parameters
			isExact = false;

			// no new upper bound was generated
			return INF;
		}

		// go to next layer
		layer++;
	}

	// take info from terminal node
	assert( node_pool.size() > 0 );
	const Node* terminal = node_pool.begin()->second;

	int ub = terminal->longest_path;
	isExact = terminal->exact;

	delete terminal;

	// if last node is exact, BDD is exact: update lower bound
	if (isExact) {
		updateLocalLB(ub, true);
	}

	// update upper bound of branch nodes
	for (vector<BranchNode*>::iterator br = localBranchNodes.begin(); br != localBranchNodes.end(); ++br) {
		(*br)->relax_ub = ub;
		estimateUpperBound(*br);
	}

	return ub;
}


//
// Merge nodes in a layer to meet maximum allowed width
//
void IndepSetBDD::mergeLayer(int layer, vector<Node*> &nodes_layer) {

	sort(nodes_layer.begin(), nodes_layer.end(), CompareNodesLongestPath());

	// All nodes will be merged to the one with index "width-1", which will
	// be denoted here by central node
	Node* central_node = nodes_layer[maxWidth-1];

	// If central node is exact, we add it to the branch node pool
	// (since it will be maintained in the BDD, we have to make a copy)
	if( central_node->exact ) {
		addBranchNode(central_node);
		central_node->exact = false;
	}

	// Merge nodes into the central node state (or simply central state)

	State* central_state = &( central_node->state );
	for (vector<Node*>::iterator node = nodes_layer.begin() + maxWidth; node != nodes_layer.end(); ++node) {

		// update state
		(*central_state) |= (*node)->state;

		// if any of the remaining nodes to merge is exact, we have to add it to the branch pool as well.
		if ((*node)->exact) {
			addBranchNode((*node));
		}

		// delete node
		delete (*node);
	}

	// resize node layer vector
	nodes_layer.resize(maxWidth);

	// Check if there are any other nodes which are equivalent to the central node

	Node* node;
	for (int i = 0; i <= maxWidth-2; ++i) {
		node = nodes_layer[i];

		// check if this state already exists in layer nodes
		if( node->state == (*central_state) ) {

			// notice that, if node exists, we do not need to update
			// the costs because the vector is already ordered

			// if existing node is exact, we have to add it to the branch pool
			// (since it will be maintained in the BDD, we have to make a copy)
			if (node->exact) {
				addBranchNode(node);
				node->exact = false;
			}

			// delete the last node (i.e., the central one)
			delete nodes_layer.back();

			// remove it from queue
			nodes_layer.pop_back();
			break;
		}
	}
}


//
// From min-in-state ordering: choose vertex that participates in the least number of states
// and remove it from the active vertex list
//
int IndepSetBDD::choose_next_vertex_min_size_next_layer() {
	int sel_index = 0;
	for (size_t i = 1; i < active_vertices.size(); ++i) {

		if (in_state_counter[active_vertices[i]] != 0 && in_state_counter[active_vertices[i]] < in_state_counter[active_vertices[sel_index]]) {
			sel_index = i;

		} else if (in_state_counter[active_vertices[i]] != 0 && (in_state_counter[active_vertices[i]] == in_state_counter[active_vertices[sel_index]]) && active_vertices[i] < active_vertices[sel_index]) {
			// lexicographic tie breaking
			sel_index = i;
		}
	}

	// TODO: check this case!!!
	if (in_state_counter[active_vertices[sel_index]] == 0) {

//		cout << "[BDD " << x10_placeID << "] POOL SIZE = " << node_pool.size() << endl;
//		cout << "State = { ";
//		for (int v = (*(node_pool.begin()->first)).find_first(); v != state_end; v = (*(node_pool.begin()->first)).find_next(v) ) {
//			cout << v << " ";
//		}
//		cout << "}" << endl;
//
//		cout << "Active vertices = { ";
//		for (vector<int>::iterator it = active_vertices.begin(); it != active_vertices.end(); ++it) {
//			cout << *it << " (counter = " << in_state_counter[*it] << ")";
//		}
//		cout << "}" << endl;
//
//		exit(1);
	}

	//assert( in_state_counter[active_vertices[sel_index]] > 0 );

	int sel_vertex = active_vertices[sel_index];
	active_vertices[sel_index] = active_vertices.back();
	active_vertices.pop_back();
	return sel_vertex;
}

//
// Construct decision diagram based on a state
//
// A state is given by (UB, longest_path, [v1,...,v_n]), where
//
//		** UB: upper bound on the solution that can be obtained by expanding this node
// 		** 'longest_path': longest path found up to the BDD node that originated state
// 		** [v_1,...,v_n] is the set of active vertices
//
//
void IndepSetBDD::x10_constructDD(const vector< vector<int> > & _states) {


	// ========================================================
	// Receive states from master, saving them as branching
	// nodes to explore
	// ========================================================

	cout << "[BDD - place = " << x10_placeID << "] Received " << _states.size() << " states to process" << endl;
	for (size_t i = 0; i < _states.size(); ++i) {
		BranchNode* branch_node = new BranchNode;

		// unserialize state
		if (_states[i].size() == 2) {
			branch_node->relax_ub = INF;
			branch_node->longest_path = 0;
			branch_node->state.resize( inst->graph->n_vertices );
			for (size_t j = 0; j < inst->graph->n_vertices; ++j) {
				branch_node->state[j] = j;
			}
		} else {
			branch_node->relax_ub = _states[i][0];
			branch_node->longest_path = _states[i][1];

			branch_node->state.resize(_states[i].size()-2);
			for (size_t j = 2; j < _states[i].size(); ++j) {
				branch_node->state[j-2] = _states[i][j];
			}
		}

		// add to branching nodes
		branchesToExplore.push(branch_node);
	}


	// ========================================================
	// Explore branching nodes
	// TODO: perhaps pool size and maximum to send dynamically
	// 		 determined?
	// ========================================================

	//long int 	maxLocalPoolSize 	= 1000;		// maximum local pool size
	//long int    maxToSend			= 20000;
	//double		fractionToSend 	= 0.15;		// fraction of nodes to send

	// to prevent local pool size
	//maxLocalPoolSize = 0;
	//fractionToSend = 1.0;


	// maximum pool size: if pool exceeds this limit,
	// then no more nodes are added to it
	const long int maxPoolSize = 1000;
	bool           emptyPool   = false;
	// fraction of generated nodes to keep locally
	//const double   fractionToKeep = 0.25;
	const double   fractionToKeep = MIN(1.0, MAX(0.1, _states.size()/100.0));

	vector<BranchNode*> nodesToEmigrate;
	while( !branchesToExplore.empty() ) {

		// get next branching node
		BranchNode* branchNode = branchesToExplore.top();
		branchesToExplore.pop();

		// explore node if better than current lower bound
		if (branchNode->relax_ub > bestLB) {

			// explore node
			int relaxedUB = generateRelaxation(branchNode);

			if (!isExact) {

				// =========================================================
				// run primal heuristic
				// TODO: generate restriction only every 'm' iterations
				// =========================================================

				// generate restriction
				int relaxedLB = generateRestriction(branchNode);
				if (relaxedLB > bestLB) {
					cout << "\n[BDD - place = " << x10_placeID << "] Primal heuristic improved lower bound: ";
					cout << "old = " << bestLB;
					cout << " - new = " << relaxedLB << endl;
				}
				updateLocalLB(relaxedLB, true);


				// =========================================================
				// emigrate nodes at local pool
				// (obs.: commented out because not so efficient...)
				// =========================================================

				if (!emptyPool) {
					const int numNodesLocal = (int) (fractionToKeep * localBranchNodes.size());

					if (numNodesLocal > 0) {
	//					cout << "[BDD - place = " << x10_placeID << "] Emigrating " << (localBranchNodes.size() - numNodesLocal);
	//					cout << " out of " << localBranchNodes.size() << " nodes";
	//					cout << endl;

						// sort local list of nodes in ascending order of UB
						sort(localBranchNodes.begin(), localBranchNodes.end(), BranchNodeComparatorUBDescending());

						// --- nodes to process locally ---
						for (int i = 0; i < numNodesLocal; ++i) {
							branchesToExplore.push( localBranchNodes.back() );
							localBranchNodes.pop_back();
						}
					}
				}
				// --- nodes to emigrate are the remaining ones in localBranchNodes ---
				emigrateBranchNodes(localBranchNodes);
			}
			else {
				// See if there are any incoming bounds if control hasn't been
				// given to X10 for a while
				// TODO: do not do this every iteration
				x10_callback->x10_step();
			}
		}

		// delete node
		delete branchNode;

		// If upper bound was updated, send it to other workers
		if (isLBUpdated) {
			//cout << "[BDD - place = " << x10_placeID << "] Sending bound: lb = " << bestLB << endl;

			x10_callback->x10_sendLB(bestLB);
			isLBUpdated = false;
		}

		// If branch size exceeds limit, pool must be emptied
		if (branchesToExplore.size() > maxPoolSize) {
			emptyPool = true;
		}

		// =========================================================
		// Emigrate nodes
		// TODO: send the best or worst nodes? random choose?
		// =========================================================

//		long int totalNodes = branchesToExplore.size() + localBranchNodes.size();
//		if (totalNodes >= maxLocalPoolSize) {
//
//
//			// we first add all elements to local branch nodes and sort
//			// them in ascending order of upper bound. The nodes
//			// with best upper bound will then be processed locally, while the
//			// remaining ones will emigrate to the master
//			//
//			// TODO: this is quite inefficient
//			//
//
//			// compute number of nodes to stay in local pool
//
//			long int numNodesToStay = (1.0 - fractionToSend) * totalNodes;
//
//			// if some nodes stay, emigrate the nodes with the best upper bound.
//			// otherwise, persist the nodes in the local queue and just send the
//			// newly generated in the branching pool (this is a heuristic)
//
//			if (numNodesToStay > 0) {
//
//				// add elements to send in local branch nodes
//				while (!branchesToExplore.empty()) {
//					localBranchNodes.push_back( branchesToExplore.top() );
//					branchesToExplore.pop();
//				}
//
//
//				nodesToEmigrate = localBranchNodes;
//				localBranchNodes.clear();
//				while (!nodesToEmigrate.empty()) {
//					if (rand()%100 <= fractionToSend * 100) {
//						localBranchNodes.push_back( nodesToEmigrate.back() );
//					} else {
//						branchesToExplore.push( nodesToEmigrate.back() );
//					}
//					nodesToEmigrate.pop_back();
//				}
//
////				// sort them in ascending order of upper bound
////				sort(localBranchNodes.begin(), localBranchNodes.end(), BranchNodeComparatorUB() );
////
////				// add nodes back to local pool
////				for (long int i = 0; i < numNodesToStay; ++i) {
////					branchesToExplore.push( localBranchNodes.back() );
////					localBranchNodes.pop_back();
////				}
////
////				cout << "[BDD - place = " << x10_placeID << "] Emigrating " << localBranchNodes.size();
////				cout << " out of " << totalNodes << " nodes";
////				cout << " (LB = " << bestLB << " - UB = " << branchesToExplore.top()->relax_ub << ")";
////				cout << endl;
//
//				// increase pool size
//				//maxLocalPoolSize *= 2;
//			}
//
//			if (localBranchNodes.size() > 0) {
//				// emigrate nodes in local branching
//				emigrateBranchNodes(localBranchNodes);
//			}
//
//
//
//		} else {
//
//			// just add elements local branching pool to queue
//			for (vector<BranchNode*>::iterator it = localBranchNodes.begin(); it != localBranchNodes.end(); ++it) {
//				branchesToExplore.push( *it );
//			}
//			localBranchNodes.clear();
//		}
//
////		// adjust local pool
////		if (branchesToExplore.size() < maxLocalPoolSize/4) {
////			maxLocalPoolSize /= 2;
////		}

	}
}


//
// Generate MDD restriction. Returns lower bound.
//
int IndepSetBDD::generateRestriction(const int initial_lp) {

	// ---------------------------------------
	// 1. Initialization
	// ---------------------------------------

	// Setup variable width
	maxWidth = MAX(10, 1000.0 / active_vertices.size());
	//cout << "width = " << maxWidth << endl;

	// initialize orderings
	if( var_ordering->order_type == MinState ) {
		for( vector<int>::iterator v = active_vertices.begin();
				v != active_vertices.end(); ++v )
		{
			in_state_counter[*v] = 1;
		}

	} else if ( var_ordering->order_type == RootOrder ) {
		ComparatorAuxIntVectorDescending comp(root_ordering);
		sort( active_vertices.begin(), active_vertices.end(), comp );

	} else {
		cout << "Order undefined" << endl;
		exit(0);
	}


	// ---------------------------------------
	// 2. Restriction
	// ---------------------------------------

	// create root node state
	State root_state;
	root_state.resize( inst->graph->n_vertices, false );
	for (vector<int>::const_iterator it = active_vertices.begin(); it != active_vertices.end(); ++it) {
		root_state.set(*it, true);
	}

	// create initial BDD node
	Node* initial_node = new Node(root_state, initial_lp, true);
	node_pool.clear();
	node_pool[ &(root_state) ] = initial_node;

	BDDNodePool::iterator node_it, existing_node_it;
	Node* node;

	// restriction control variables
	int current_vertex;
	int layer = 0;
	const int num_active_vertices = active_vertices.size();

	while ( layer < num_active_vertices ) {

		// select next vertex
		if( var_ordering->order_type == MinState ) {
			current_vertex = choose_next_vertex_min_size_next_layer();

		} else if ( var_ordering->order_type == RootOrder ) {
			current_vertex = active_vertices.back();
			active_vertices.pop_back();

		} else {
			exit(0);
			//current_vertex = var_ordering->vertex_in_layer(NULL, layer);
		}

		// Take nodes from the pool that have the current vertex in their state
		nodes_layer.clear();
		node_it = node_pool.begin();
		while( node_it != node_pool.end() )	{

			if( node_it->second->state[current_vertex] ) {

				// if var ordering is min-in-state, decrement active list
				if( var_ordering->order_type == MinState ) {
					for( vector<int>::iterator v = active_vertices.begin();
							v != active_vertices.end();
							++v )
					{
						if( (*node_it->first)[*v] ) {
							in_state_counter[*v]--;
						}
					}
				}

				// move from the pool to current layer list
				nodes_layer.push_back(node_it->second);
				node_pool.erase(node_it++);

			} else {
				++node_it;
			}
		}

//		// PRINT LAYER
//		cout << "Layer " << layer << " - current vertex: " << current_vertex;
//		cout << " - pool size: " << node_pool.size();
//		cout << " - before merge: " << nodes_layer.size();
//		cout << " - total: " << node_pool.size() + nodes_layer.size();
//		cout << endl;

		// 2. Restriction
		if( maxWidth != INF && (int)nodes_layer.size() > maxWidth ) {
			restrictLayer(layer, nodes_layer);
		}


		// 3. Branching
		Node* branch_node;
		for( vector<Node*>::iterator it = nodes_layer.begin(); it != nodes_layer.end(); ++it ) {

			branch_node = (*it);
			branch_node->state.set(current_vertex, false);

			// --------------- One arc ---------------------

			node = new Node(branch_node->state,
					branch_node->longest_path + inst->graph->weights[current_vertex],
					branch_node->exact);

			// remove adjacency vertices
			node->state &= inst->adj_mask_compl[current_vertex];

			existing_node_it = node_pool.find( &(node->state) );
			if( existing_node_it != node_pool.end() ) {

				// node already exists in the pool: update node match
				existing_node_it->second->longest_path = MAX(node->longest_path,
						existing_node_it->second->longest_path);
				delete node;

			} else {
				node_pool[ &(node->state)] = node;

				// update eligibility list
				if( var_ordering->order_type == MinState ) {
					for( vector<int>::iterator v = active_vertices.begin();
							v != active_vertices.end();
							++v )
					{
						if( (node->state)[*v] ) {
							in_state_counter[*v]++;
						}
					}
				}
			}


			// --------------- Zero arc ---------------------

			node = branch_node;  // we can reuse the branch node!

			existing_node_it = node_pool.find( &(node->state) );
			if( existing_node_it != node_pool.end() ) {

				// node already exists in the pool: update node match
				existing_node_it->second->longest_path = MAX(node->longest_path,
						existing_node_it->second->longest_path);

				delete node;

			} else {
				node_pool[ &(node->state)] = node;

				// update eligibility list
				if( var_ordering->order_type == MinState ) {
					for( vector<int>::iterator v = active_vertices.begin();
							v != active_vertices.end();
							++v )
					{
						if( (node->state)[*v] ) {
							in_state_counter[*v]++;
						}
					}
				}
			}

		}

		// go to next layer
		layer++;
	}

	// take terminal node
	Node* terminal = node_pool.begin()->second;
	long int lb = terminal->longest_path;

	// delete terminal
	delete terminal;
	return lb;
}

//
// Reduce the size of the layer
//
void IndepSetBDD::restrictLayer(int layer, vector<Node*> &nodes_layer) {
	sort(nodes_layer.begin(), nodes_layer.end(), CompareNodesLongestPath());
	for( vector<Node*>::iterator node = nodes_layer.begin()+maxWidth;
			node != nodes_layer.end(); ++node)
	{
		delete (*node);
	}
	nodes_layer.resize(maxWidth);
}



