/*
 * ===============================================================
 * IndepSetBDD - Definition class
 * ===============================================================
 */

#ifndef INDEPSETBDD_HPP_
#define INDEPSETBDD_HPP_

#include <map>
#include <queue>
#include <deque>

#include "indepset_orderings.hpp"
#include "indepset_instance.hpp"
#include "util.hpp"

#include "../../../../DDX10_Base.h"

#define EPS 1e-8

using namespace std;


//
// Parameters
// TODO: include this in a parameter struct
//
#define ROOT_WIDTH	100				// width for initial relaxation



// -------------------------------------------------------------------
// Branching Node
// -------------------------------------------------------------------

struct BranchNode {
	vector<int>		state;				// state to be explored
	int   			longest_path;		// node longest path
	int   			relax_ub;           // upper bound obtained for the BDD where node was created

	//
	// Constructor from a BDD node
	//
	BranchNode(Node* _node, double _longest_path)
	:	longest_path(_longest_path), relax_ub(0)
	{
		const State& state_v = _node->state;
		state.reserve( state_v.count() );
		for (int v = state_v.find_first(); v != state_end; v = state_v.find_next(v)) {
			state.push_back(v);
		}
	}

	//
	// Constructor from a state vector
	//
	BranchNode(const vector<int>& _state, double _longest_path)
	:	state(_state), longest_path(_longest_path), relax_ub(0)
	{
	}

	//
	// Empty constructor
	//
	BranchNode() { }
};


//
// Branch node comparator by relaxed UB
//
struct BranchNodeComparatorUB {
	BranchNodeComparatorUB() { }
	bool operator()(const BranchNode* b1, const BranchNode* b2) {
		if (b1->relax_ub == b2->relax_ub) {
			if (b1->state.size() == b2->state.size()) {
				return b1->longest_path < b2->longest_path;
			}
			return b1->state.size() < b2->state.size();
		}
		return b1->relax_ub < b2->relax_ub;
	}
};


//
// Branch node comparator by relaxed UB (descending)
//
struct BranchNodeComparatorUBDescending {
	BranchNodeComparatorUBDescending() { }
	bool operator()(const BranchNode* b1, const BranchNode* b2) {
		if (b1->relax_ub == b2->relax_ub) {
			if (b1->state.size() == b2->state.size()) {
				return b1->longest_path > b2->longest_path;
			}
			return b1->state.size() > b2->state.size();
		}
		return b1->relax_ub > b2->relax_ub;
	}
};

//
// Queue of branch nodes ordered by relax ub
//
typedef priority_queue<BranchNode*, vector<BranchNode*>, BranchNodeComparatorUB> BranchNodeQueue;


// -------------------------------------------------------------------
// Independent Set BDD: Relaxation / Restriction generator
// -------------------------------------------------------------------

// Independent set solver class
class IndepSetBDD : public DDX10_Base {

public:

	//
	// Constructor given a DIMACS instance filename
	//
	// IndepSetBDD(char* filename, int _maxWidth)
	// DDX10:
	IndepSetBDD(const int _placeID,
				const int _ddWidth,
				const string & _instanceName,
				DDX10_Callback * const _cb);

	// Generate relaxation.
	int generateRelaxation(const int initial_weight);

	// Generate restriction
	int generateRestriction(const int initial_weight);

	// Generate relaxation from a branch node.
	int generateRelaxation(const BranchNode* branchNode);

	// Generate restriction from a branch node.
	int generateRestriction(const BranchNode* branchNode);


	// Update local lower bound, indicating whether it was found locally or remotely
	void updateLocalLB(int _lb, bool isLocal = false);

	// Verify if relaxed BDD is exact
	bool isBDDExact() 				{ return isExact; }

	// Get best lower bound found
	int getBestLB() 		  		{ return bestLB; }

	// DDX10: implementation of virtual methods of the DDX10_Base base class
	void   x10_constructDD(const vector< vector<int> > & states);
	void   x10_updateLB(const int lb);
	void   x10_updateUB(const int ub) { }
	int    x10_nVars(void) const { return inst->graph->n_vertices; }
	void   x10_printSolution(void) { }
	void   x10_printStats(void) { }

	// Public data members
	IndepSetInst* 			inst;				// independent set instance
	vector<int> 			vertex_in_layer;  	// which vertex a layer corresponds to
	Node*					root_node;		  	// relaxation root node

private:

	// ----------------------------------------------
	// Attributes
	// ----------------------------------------------

	BDDNodePool 		node_pool;		  		// list of nodes to explore

	vector<int>			active_vertices;  		// active vertices in the relaxation
	vector<Node*> 		nodes_layer;	  		// nodes in a layer

	int* 				in_state_counter;  		// min-in-state vertex ordering parameters
	IS_Ordering* 		var_ordering; 			// BDD Vertex ordering

	int 				maxWidth;  				// max allowed width
	double 				bestLB;				// best lower bound
	bool				isExact;				// if BDD is exact

	bool				isLBUpdated;			// if lower bound has been updated

	vector<BranchNode*> branch_nodes; 			// BDD nodes that require branching
	vector<int>			root_ordering;			// relative ordering of each vertex


	BranchNodeQueue		branchesToExplore;		// nodes that require exploration
	vector<BranchNode*> localBranchNodes; 		// new branch nodes generated during relaxation


	// Auxiliary parameters
	CompareNodesLongestPath bn_comparator;


	// ----------------------------------------------
	// Functions
	// ----------------------------------------------

	// Choose vertex
	int choose_next_vertex_min_size_next_layer();

	// Merge two nodes
	void merge(Node* nodeA, Node* nodeB);

	// Merge layer
	void mergeLayer(int layer, vector<Node*> &nodes_layer);

	// Restrict layer
	void restrictLayer(int layer, vector<Node*> &nodes_layer);

	// Add new local node to be explored later
	void addBranchNode(Node* _node);

	// Estimate upper bound of a branch node
	void estimateUpperBound(BranchNode* branchNode);

	// Emigrate nodes
	void emigrateBranchNodes(vector<BranchNode*>& branchNodes);
};


// -------------- Utilities ----------------

struct CompareOrderDescending {
	vector<int> &order;
	CompareOrderDescending(vector<int>& _order) : order(_order) { }
	bool operator()(const int a, const int b) {
		return order[a] > order[b];
	}
};


// ----------------------------------------------------
// Inline implementations
// ----------------------------------------------------

//
// Merge node A into node B. Node A is always deleted.
//
inline void IndepSetBDD::merge(Node* nodeA, Node* nodeB) {
	assert( (nodeA->state) == (nodeB->state) );

	if ( (nodeA->exact == nodeB->exact) ) {

		// Both nodes are either relaxed or exact
		nodeB->longest_path = MAX(nodeA->longest_path, nodeB->longest_path);

	} else if (nodeA->exact) {

		// Node A is exact, B is relaxed
		branch_nodes.push_back( new BranchNode(nodeA, nodeA->longest_path) );
		nodeB->longest_path = MAX(nodeA->longest_path, nodeB->longest_path);

	} else if (nodeB->exact) {

		// Node B is exact, A is relaxed
		branch_nodes.push_back( new BranchNode(nodeB, nodeB->longest_path));

		nodeB->exact = false;
		nodeB->longest_path = MAX(nodeA->longest_path, nodeB->longest_path);
	}

	delete nodeA;
}

//
// Update lower bound, indicating whether it was found locally or remotely. If locally,
// then mark that bound was updated to send it to other nodes
//
inline void IndepSetBDD::updateLocalLB(int _lb, bool isLocal) {
	if (_lb > bestLB) {
		bestLB = _lb;
		if (isLocal) {
			isLBUpdated = true;
		}

		//cout << "[BDD - place = " << x10_placeID << "] New lower bound = " << bestLB;
		//cout << (isLocal ? " (found locally)" : "");
		//cout << endl;
	}
}


//
// Add new local node to be explored
//
inline void IndepSetBDD::addBranchNode(Node* _node) {
	localBranchNodes.push_back(	new BranchNode(_node, _node->longest_path) );
}


//
// Update best lower bound
//
inline void IndepSetBDD::x10_updateLB(const int lb) {
	updateLocalLB(lb, false);
}


//
// Heustic to estimate upper bound of a branch node
//
inline void IndepSetBDD::estimateUpperBound(BranchNode* branchNode) {
	branchNode->relax_ub = MIN(	branchNode->relax_ub,
								branchNode->longest_path + branchNode->state.size());
}


//
// Generate relaxation from a branch node.
//
inline int IndepSetBDD::generateRelaxation(const BranchNode* branchNode) {
	active_vertices = branchNode->state;
	return generateRelaxation( branchNode->longest_path );
}


//
// Generate restriction from a branch node.
//
inline int IndepSetBDD::generateRestriction(const BranchNode* branchNode) {
	active_vertices = branchNode->state;
	return generateRestriction( branchNode->longest_path );
}

//
// Emigrate nodes
//
inline void IndepSetBDD::emigrateBranchNodes(vector<BranchNode*>& branchNodes) {
	if (branchNodes.empty())
		return;

	vector< vector<int> > states;
	for (vector<BranchNode*>::iterator st = branchNodes.begin(); st != branchNodes.end(); ++st) {
		BranchNode* branchNode = (*st);
		if (branchNode->relax_ub > bestLB) {
			states.push_back( vector<int>() );
			states.back().reserve( branchNode->state.size() + 2 );
			states.back().push_back( branchNode->relax_ub );
			states.back().push_back( branchNode->longest_path );
			states.back().insert( states.back().end(), branchNode->state.begin(), branchNode->state.end() );
		}
		delete branchNode;
	}
	branchNodes.clear();

	// emigrate nodes: send states
	//cout << "[BDD - place = " << x10_placeID << "] Sending " << states.size() << " states..." << endl;
	x10_callback->x10_sendStates(states);
}



#endif /* INDEPSETBDD_HPP_ */
