// -----------------------------------------------------------------
// Licensed Materials - Property of IBM
//
// DDX10
//
// (C) Copyright IBM Corporation 2013.  All Rights Reserved.
//
// Authors: Andre Cire, Ashish Sabharwal, Horst Samulowitz,
//          Vijay Saraswat
// -----------------------------------------------------------------


import x10.io.Console;
import x10.io.File;
import x10.io.FileWriter;
import x10.io.Printer;
import x10.compiler.Native;
import x10.compiler.NativeRep;
import x10.compiler.NativeCPPOutputFile;
import x10.compiler.NativeCPPCompilationUnit;
import x10.util.Pair;

import com.ibm.ddx10.StatePool;


public class DDX10 {
	
	// base class for all supported Decision Diagram solvers
	@NativeRep("c++", "DDX10__DD*", "DDX10__DD", null)
	@NativeCPPOutputFile("DDX10__DD.h")
	@NativeCPPCompilationUnit("DDX10__DD.cc")
	@NativeCPPOutputFile("DDX10_Base.h")
	
	static class DD {
		public def this():DD {}
		
		@Native("c++", "(#0)->attachCallbacks((#1),(#2),(#3))")
		public def attachCallbacks((Rail[Int])=>void, (Int)=>void, (Int)=>void):void {}
		
		@Native("c++", "(#0)->constructDD((#1))")
		public def constructDD(Rail[Int]):void {}
		
		@Native("c++", "(#0)->updateLB((#1))")
		public def updateLB(Int):void {}
		
		@Native("c++", "(#0)->updateUB((#1))")
		public def updateUB(Int):void {}
		
		@Native("c++", "(#0)->kill()")
		public def kill():void {}
		
		@Native("c++", "(#0)->printStats()")
		public def printStats():void {}
		
		@Native("c++", "(#0)->printResults((#1))")
		public def printResults(Boolean):void {}
	}
	
	/////////////////  BEGIN DD solver declarations //////////////////
	
	// derived class for IndepSetBDD
	@NativeRep("c++", "DDX10__DD_IndepSetBDD*", "DDX10__DD_IndepSetBDD", null)
	@NativeCPPOutputFile("DD/c++_x10_wrappers/DDX10__DD_IndepSetBDD.h")
	@NativeCPPCompilationUnit("DD/c++_x10_wrappers/DDX10__DD_IndepSetBDD.cc")
	@NativeCPPOutputFile("DD/c++/IndepSetBDD/src/bdd.hpp")
	@NativeCPPOutputFile("DD/c++/IndepSetBDD/src/indepset_bdd.hpp")
	@NativeCPPOutputFile("DD/c++/IndepSetBDD/src/indepset_instance.hpp")
	@NativeCPPOutputFile("DD/c++/IndepSetBDD/src/indepset_orderings.hpp")
	@NativeCPPOutputFile("DD/c++/IndepSetBDD/src/util.hpp")
	
	static class DD_IndepSetBDD extends DD {
		public def this (placeID:Int, ddWidth:Int, filename:String) : DD_IndepSetBDD { }
	}
	
	/////////////////  END DD solver declarations //////////////////
	
	// the main DD solver class containing instances of the desired DD solver type
	static class DDSolver {
		var dd:DD;   /** a Decision Diagram object associated with this solver */
		val statePool = new StatePool();   /** a state pool */
		var isMaximization:Boolean;
		
		public def this(ddType:String, ddWidth:Int, filename:String, statePoolType:String) {
			this(ddType, ddWidth, filename, statePoolType, 0.0);
		}
		public def this(ddType:String, ddWidth:Int, filename:String,
				        statePoolType:String, statePoolRandFrac:Double) {
			Console.OUT.println("c " + here
					+ ": ddType " + ddType
					+ ", statePoolType " + statePoolType
					+ ", randFrac " + statePoolRandFrac);
			if (ddType.equals("IndepSetBDD")) {
				dd = new DD_IndepSetBDD(here.id, ddWidth, filename);
				isMaximization = true;
				statePool.createPool(statePoolType, isMaximization, statePoolRandFrac);
			}
			else {
				Console.OUT.println("c WARNING: Unknown DD type " + ddType + ". Ignoring it!");
				dd = null;
			}
		}
	}
	
	
	// a boolean array to keep track of which places are currently actively working
	static val placeActive <: Rail[Boolean] = new Rail[Boolean](Place.MAX_PLACES, false);
	
	// main method
	public static def main (args:Array[String](1)) {
		Console.OUT.println("c ==== DDX10: Parallel Decision Diagram Solver ====");
		// Store start time
		val lStartTime = System.nanoTime();
		
		// parse command line
		if (args.size < 3 || args.size > 5) {
			Console.ERR.println("USAGE: DDX10 ddType ddWidth instanceName [timeLimit] [statsFile]");
			return;
		}
		val ddType    = args(0);
		val ddWidth   = Int.parse(args(1));
		val filename  = args(2);
		val timeLimit = (args.size >= 4 ? Int.parse(args(3)) : -1) * 1000000000.0;   // wall clock time limit (in nanoseconds)
		val statsFile = (args.size >= 5 ? args(4) : "stats_ddx10.txt"); 
		Console.OUT.println("c DDX10 parameters:");
		Console.OUT.println("c   DD type                   : " + ddType);
		Console.OUT.println("c   DD width                  : " + ddWidth);
		Console.OUT.println("c   wall clock timelimit      : " + Int.operator_as(timeLimit / 1000000000.0) + " seconds");
		Console.OUT.println("c   stats file                : " + statsFile);
		Console.OUT.println("c   instance filename         : " + filename);
		// TODO: fix temporary hack: if nplaces>1 then do not use DDSolver at master
		val nsolvers = (Place.MAX_PLACES == 1 ? 1 : Place.MAX_PLACES - 1);
		if (Place.MAX_PLACES == 1)
			Console.OUT.println("c   num solvers               : 1 (also the master)");
		else
			Console.OUT.println("c   num solvers               : " + nsolvers + " + master");
		
		// create a solver at each place
		val everyone = Dist.makeUnique();
		val s_handle = PlaceLocalHandle.make[DDSolver](everyone,
				// () => new DDSolver(ddType, ddWidth, filename, "RandomPool")
				// () => new DDSolver(ddType, ddWidth, filename, "StackPool")
				// () => new DDSolver(ddType, ddWidth, filename, "QueuePool")
				// () => new DDSolver(ddType, ddWidth, filename, "OccasionallySortedQueuePool")
				// () => new DDSolver(ddType, ddWidth, filename, "PQueuePool", 0.0)
				() => new DDSolver(ddType, ddWidth, filename, "PQueueDFSPool", 0.0)
				);
		
		// place ID of the master process that handles StatePool; this place
		// also runs a regular DD solver like the rest of the places, responding
		// to getState() requests whenever it handles incoming messages
		val master_id = here.id();
		val master    = Place.place(master_id);
		Console.OUT.println("c " + here + ": starting master process for StatePool");
		
		// launch a DD solver at each place; wait for all solvers to finish 
		// (by solving or by being killed)
		
		finish ateach (i in everyone) {
			try {
				// define a callback to the master process for adding new states to the pool		
				val cb_sendStates = (_states:Rail[Int]) => {
					if (here.id == master_id)
						s_handle().statePool.addStates(_states);											
					else
						at (master) s_handle().statePool.addStates(_states);						
				};
				
				// define a callback to send a lower bound to every other place and to the state pool
				val not_here = new Rail[Place](Place.MAX_PLACES-1, 
						(i:Int)=>Place.place(i<here.id ? i : i+1));
				val cb_sendLB = (_lb:Int) => {
					for (j in not_here.values()) at (j) async {
						//Console.OUT.println("c " + here + " updating lower bound received from afar");
						s_handle().dd.updateLB(_lb);
					}
					if (here.id == master_id)
						s_handle().statePool.updateLB(_lb);
					else
						at (master) s_handle().statePool.updateLB(_lb);
				};
				
				// define a callback to send an upper bound to every other place and to the state pool
				val cb_sendUB = (_ub:Int) => {
					for (j in not_here.values()) at (j) async {
						Console.OUT.println("c " + here + " updating upper bound received from afar");
						s_handle().dd.updateUB(_ub);
					}
					if (here.id == master_id)
						s_handle().statePool.updateUB(_ub);
					else
						at (master) s_handle().statePool.updateUB(_ub);
				};
				
				// attach callbacks to the DD solver
				s_handle().dd.attachCallbacks(cb_sendStates, cb_sendLB, cb_sendUB);
				
				// start the DD solver
				// TODO: fix temporary hack: if nplaces>1 then do not use DDSolver at master
				if (Place.MAX_PLACES > 1 && here.id == master_id)
					Console.OUT.println("c " + here + ": MASTER NOT STARTING DD solver (TEMPORARY)");
				else
					Console.OUT.println("c " + here + ": starting DD solver");
				val myStartTime = System.nanoTime();
				
				// Decay sleep time
				//val dSleepDecrease        = 0.5;
				//var nSleepIntervalCurrent = 1000;
					
				// the perpetual while loop...
				var done:Boolean = false;
				val maxiter = 100000000;
				var iter:ULong = 0;
				
				//while (iter++ < maxiter && !done) 
				while (!done) {
					iter++;
				
					// TODO: fix temporary hack: if nplaces>1 then do not use DDSolver at master
					if (Place.MAX_PLACES > 1 && here.id == master_id)
						break;
					
					//!!if (iter % 10 == 0) {	
					//  Console.OUT.print(here);		
					//  Console.OUT.print(" Size of State Pool: ");				
					//  Console.OUT.print(at (master) s_handle().statePool.getPoolSize());
					//  Console.OUT.print(" - LB = "); 
					// Console.OUT.println(at (master) s_handle().statePool.getLB());
					//}
					
					// check if time limit was reached
					//val tl = (System.nanoTime() - lStartTime);
					val tl = (System.nanoTime() - myStartTime);
					if ( timeLimit > 0 && tl > timeLimit ) {
						Console.OUT.println(here + " aborting due to time limit");
						break;
					}
																									
					if (iter % 10UL == 0UL) {
						if (s_handle().isMaximization) {
							val LB = at (master) s_handle().statePool.getLB();
							val poolUB = at (master) s_handle().statePool.getPoolUB(true);
							val poolSize = at (master) s_handle().statePool.getPoolSize();
							Console.OUT.println(here
									+ " : Iter = " + iter
									+ " | LB = " + LB
									+ " | PoolUB = " + poolUB
									+ " | PoolSize = " + poolSize
									+ " | Timestamp = " + tl / 1000000000.0
									);
						}
						else {
							val poolLB = at (master) s_handle().statePool.getPoolLB(true);
							val UB = at (master) s_handle().statePool.getUB();
							val poolSize = at (master) s_handle().statePool.getPoolSize();
							Console.OUT.println(here
									+ " : Iter = " + iter
									+ " | poolLB = " + poolLB
									+ " | UB = " + UB
									+ " | PoolSize = " + poolSize
									+ " | Timestamp = " + tl / 1000000000.0
									);
						}
					}
															
					val retValue = (here.id == master_id) ?	s_handle().statePool.getStates(-1) : at (master) s_handle().statePool.getStates(-1);
					val nStates = retValue.first;
					val states  = retValue.second;
					//Console.OUT.println("c " + here + ": got " + nStates + " states");
					switch (nStates) {
					case -1:
						// done! no more states in the pool and no more outstanding jobs. 
						// this means the lower bound is the optimum solution
						done = true;
						at (master) {
							val opt = s_handle().statePool.getLB();
							s_handle().statePool.updateUB(opt);
						}
						break;
					case 0:
						// sleep for a bit and try again
						Console.OUT.println(here + " sleeping");
						System.sleep(200);  // sleep for 1 second (specified in milliseconds)
						//nSleepIntervalCurrent = max(nSleepIntervalCurrent * dSleepDecrease, 50.0);
						break;
					default:
						// got states; call the DD construction routine on these
						s_handle().dd.constructDD(states);
						// done creating DDs; decrement outstanding tasks counter
						if (here.id == master_id)
							s_handle().statePool.decrOutstandingTasks(nStates);
						else
							at (master) s_handle().statePool.decrOutstandingTasks(nStates);
					}					
				}
				
				//s_handle().dd.printResults(false);

				Console.OUT.flush();
				
				// // process the results
				// if (result != 0) {   // 0: unsolved, 1: satisfiable; -1: unsatisfiable
				// 	// the instance was solved here!
				// 	Console.OUT.println("c " + here + ": INSTANCE SOLVED!");
				// 	// compute wall-clock time so far and print it
				// 	val myEndTime   = System.nanoTime();
				// 	val dSeconds = (myEndTime - myStartTime) / 1000000000.0;
				// 	Console.OUT.println("c Solver Wall-Clock Time       : " + dSeconds + " seconds");      
				// 	// send kill signal to all other solvers
				// 	for (j in not_here.values()) 
				// 		at (j) async s_handle().kill();
				// 	// record that this place succeeded in solving the instance
				// 	if (here.id == master_id)
				// 		placeSuccessful(master_id) = true;
				// 	else {
				// 		val myID = here.id;
				// 		at (Place.place(master_id)) placeSuccessful(myID) = true;
				// 	}
				// }
				// else {
				// 	// the instance wasn't solved; got here most likely because the solver was
				// 	// killed (or timed out or crashed internally)
				// 	Console.OUT.println("c " + here + ": aborting");
				// }
			}
			catch (exc:Exception) {
				Console.OUT.println("c " + here + ": aborted because of " + exc.toString());
                Console.OUT.flush();
			}
			
		} // end: finish
		
		assert (here.id == master_id);
		
		// compute the final upper and lower bounds; note that this relies
		// on having no outstanding tasks! also, depends on whether this
		// is a maximization problem or a minimization problem
		assert (s_handle().statePool.getNOutstandingTasks() == 0);
		val LB = (s_handle().isMaximization ? s_handle().statePool.getLB() : s_handle().statePool.getPoolLB(false));
		val UB = (s_handle().isMaximization ? s_handle().statePool.getPoolUB(false) : s_handle().statePool.getLB());

		// print statistics for an arbitrarily choosen place that ran the solver
		if (Place.MAX_PLACES == 1)
			s_handle().dd.printStats();
		else
			at (Place.place(1)) s_handle().dd.printStats();

		Console.OUT.println("c final lower bound           : " + LB);
		Console.OUT.println("c final upper bound           : " + UB);
		Console.OUT.println("c states pruned               : " + s_handle().statePool.getNStatesPruned());		
		Console.OUT.println("c states processed            : " + s_handle().statePool.getNStatesProcessed());
		Console.OUT.println("c max pool size reached       : " + s_handle().statePool.getMaxPoolSizeReached() + " states");
			
		// Get End Time
		val lEndTime = System.nanoTime();
		// Compute duration in seconds
		val dSeconds = (lEndTime - lStartTime) / 1000000000.0;
		// Output total wall clock time 
		Console.OUT.printf("c Total time till full exit   : %0.2f seconds\n", dSeconds);      		
			
		val output = new FileWriter(new File(statsFile), true);
		val stats = new Printer(output);
		stats.print(filename
				+ "\t" + nsolvers
				+ "\t" + LB
				+ "\t" + UB
				+ "\t" + s_handle().statePool.getNStatesPruned()    // Keeping this here for compatibity with old stats files
				);
		stats.printf("\t%.2f", dSeconds);
		stats.println("\t" + s_handle().statePool.getNStatesProcessed()
				+ "\t" + s_handle().statePool.getNStatesPruned()
				+ "\t" + s_handle().statePool.getMaxPoolSizeReached()
				);
		stats.flush();
		stats.close();			
	}
}


// vim: shiftwidth=4:tabstop=4
