// -----------------------------------------------------------------
// Licensed Materials - Property of IBM
//
// DDX10
//
// (C) Copyright IBM Corporation 2013.  All Rights Reserved.
//
// Authors: Andre Cire, Ashish Sabharwal, Horst Samulowitz,
//          Vijay Saraswat
// -----------------------------------------------------------------


#ifndef DDX10_Base_h
#define DDX10_Base_h

#include <iostream>
#include <vector>
#include <string>
using std::vector;
using std::string;


// these are the callback functions that will be available to the DD solver
class DDX10_Callback {
public:
    // methods that inheriting Callback classes must support;
    // currently being implemented in DDX10_DD
    virtual void           x10_step(void) const = 0;
    virtual void           x10_sendStates(const vector< vector<int> > & states) = 0;
    virtual void           x10_sendLB(const int lb) = 0;
    virtual void           x10_sendUB(const int ub) = 0;
};


// The base class that every DD class must inherit from.
// Serves three purposes:
//   A. Makes available to each solver various x10 related objects such
//      as place ID, max length of clauses to share, pointer to a callback
//      object, clause buffers, etc.
//   B. Lists several pure virtual methods that each inheriting solver is
//      expected to implement in order to interact with DDX10.x10
//   C. Provide a number of additional controls for DDX10.x10
//
class DDX10_Base {
protected:
    // default values of key parameters
	static const int    default_x10_messageProcessingInterval = 100;   // default number of decisions after which to process incoming messages

    // objects that this base class keeps track of and makes available
    // to all inheriting solvers (e.g., place ID, callback, etc.)
    const int                 x10_placeID;                // the ID of the X10 place that is running this solver
    const int                 x10_ddWidth;                // maximum width of the DD
    const string              x10_instanceName;           // name of the CNF instance being solved
    DDX10_Callback *    const x10_callback;               // the 'callback' through which this solver interacts with X10
    const int                 x10_messageProcessingInterval; // after how many decisions to call step() / probe() to incorporate incoming messages

    int  x10_whenMessagesWereProcessed;      // to keep track of the last time incoming messages were processed
    bool x10_killed;                         // whether a kill signal has been received from X10

    // internal methods that inheriting SAT solvers must support:

public:
    // constructors
    DDX10_Base(
            const int       _placeID,
            const int       _ddWidth,
            const string &  _instanceName,
            DDX10_Callback * const _cb = NULL,
            const int       _messageProcessingInterval = default_x10_messageProcessingInterval) :
        x10_placeID(_placeID),
        x10_ddWidth(_ddWidth),
        x10_instanceName(_instanceName),
        x10_callback(_cb),
        x10_messageProcessingInterval(_messageProcessingInterval),
        x10_whenMessagesWereProcessed(0),
        x10_killed(false)
    { }


    // public methods that inheriting SAT solvers must support
    virtual void   x10_constructDD(const vector< vector<int> > & states) = 0;
    virtual void   x10_updateLB(const int lb) = 0;
    virtual void   x10_updateUB(const int ub) = 0;
    virtual int    x10_nVars(void) const = 0;
    virtual void   x10_printSolution(void) = 0;
    virtual void   x10_printStats(void) = 0;

    // additional controls for DDX10.x10 or for inheriting solvers
    int    x10_getPlaceID         (void) const { return x10_placeID; }
    int    x10_getDdWidth         (void) const { return x10_ddWidth; }
    bool   x10_wasKilled          (void) const { return x10_killed; }

    void   x10_kill               (void) { x10_killed = true; }
};


///////////////////////////////
// INLINE METHODS
///////////////////////////////



#endif
