// -----------------------------------------------------------------
// Licensed Materials - Property of IBM
//
// DDX10
//
// (C) Copyright IBM Corporation 2013.  All Rights Reserved.
//
// Authors: Andre Cire, Ashish Sabharwal, Horst Samulowitz,
//          Vijay Saraswat
// -----------------------------------------------------------------


#include "x10/array/Array.h"
#include "x10/lang/String.h"

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include "DDX10__DD.h"


void CallbackStats::printStats(const DDX10_Base * const dd) const {
    cout.setf(std::ios::fixed); cout.precision(2);  // set precision to 2 decimal digits

    cout << "c == Place(" << dd->x10_getPlaceID() << ") stats ==" << endl
         << "c x10 ddWidth                 : " << dd->x10_getDdWidth() << endl
         << "c x10 incoming states         : " << numIncomingStates << endl
         << "c x10 outgoing states         : " << numOutgoingStates << endl
         << "c x10 incoming bounds         : " << numIncomingBounds << endl
         << "c x10 outgoing bounds         : " << numOutgoingBounds << endl
         << endl;
}

// type 1: implement for base class DDX10_Callback:
//   process outgoing states by calling the appropriate closure to share them
void DDX10__DD::x10_sendStates(const vector< vector<int> > & states) {
    // update x10 callback statistics
    cb_stats.incrementNumOutgoingStates(states.size());

    // convert states to an x10 array
    // first compute the total space needed
    unsigned totalSize = 0;
    for (unsigned i=0 ; i<states.size() ; i++)
    	totalSize += states[i].size() + 1;   // one additional location per state

    x10::array::Array<int>* states_ = x10::array::Array<int>::_make(totalSize);
    x10_int pos = 0;
    for (unsigned i=0 ; i<states.size() ; i++) {
        states_->FMGL(raw)[pos++] = states[i].size();
        for (unsigned j=0 ; j<states[i].size() ; j++) {
            //(*clause_)[i] = toInt(clause[i]); // original
            //(*clause_).__set(i, toInt(clause[i])); // valid?
           states_->FMGL(raw)[pos++] = states[i][j];
        }
    }

//    cout << "serialized vector: ";
//    for (unsigned k = 0; k < totalSize; ++k) {
//    	cout << states_->FMGL(raw)[k] << " ";
//    }
//    cout << endl;

    // apply the closure to the clauses with info
    // IMPORTANT NOTE: when passing control to X10 to execute the closure, X10 may
    // very well process incoming messages, including the kill() message and the
    // updateStates() message! Hence, best to call the closure at the end
    // here, when we are not actively handling incoming or outgoing buffers
    x10::lang::VoidFun_0_1<x10::array::Array<int>* >::__apply(closure1, states_);
}

// type 2: main DD construction method, to be called by DDX10.x10
void DDX10__DD::constructDD(x10::array::Array<int>* states_) {
    // de-serialize: first convert from x10 array to vector<vector<int> >
    const vector<int>      emptyVec;
    vector< vector<int> >  states;

	int i=0;
    while (i<states_->FMGL(raw)->length()) {
       	const int nextStateSize = states_->FMGL(raw)[i++];
       	states.push_back(emptyVec);
       	states.back().resize((unsigned) nextStateSize);
       	for (int j=0; j<nextStateSize; j++)
       		states.back()[j] = states_->FMGL(raw)[i++];
    }

    // push clause with info into solver's incoming buffer
    dd->x10_constructDD(states);
    // update x10 callback statistics
    cb_stats.incrementNumIncomingStates(states.size());
}

// type 2: print instance information
void DDX10__DD::printInstanceInfo(void) const {
    if (dd == NULL) {
        cout << "WARNING: instance information can be printed only after a DD object has been created." << endl;
        return;
    }
    cout << "c Instance information:" << endl
         << "c   filename                  : " << instanceName << endl
         << "c   number of variables       : " << dd->x10_nVars() << endl;
}

// type 2: print final result (solver statistics, sat/unsat, plus solution if requested)
void DDX10__DD::printResults(const bool flag_printsoln) const {
    // print result and possibly the solution, if found and asked for
//    switch (solveStatus) {
//    case -1:
//        printf("s UNSATISFIABLE\n");
//        break;
//
//    case 0:
//        printf("s UNKNOWN\n");
//        break;
//
//    case 1:
//        printf("s SATISFIABLE\n");
//        if (flag_printsoln)
//            printSolution();
//        break;
//
//    default:
//        abort();
//    }

    // print solver's native statistics
    printStats();
    // print x10 callback statistics
    cb_stats.printStats(dd);

    // flush the output, just in case the process gets killed
    cout << std::flush;
}


RTT_CC_DECLS0(DDX10__DD, "DDX10.DD", x10aux::RuntimeType::class_kind)
