// -----------------------------------------------------------------
// Licensed Materials - Property of IBM
//
// DDX10
//
// (C) Copyright IBM Corporation 2013.  All Rights Reserved.
//
// Authors: Andre Cire, Ashish Sabharwal, Horst Samulowitz,
//          Vijay Saraswat
// -----------------------------------------------------------------


#ifndef DDX10__DD_h
#define DDX10__DD_h

#include "x10/lang/VoidFun_0_1.h"
#include "x10/lang/Fun_0_0.h"
#include "x10/lang/X10Class.h"

#include "DDX10_Base.h"

#include <string>


/** A simple class to keep track of statistics related to x10 callbacks */
class CallbackStats {
protected:
    int         numIncomingStates;        // total number of incoming states
    int         numIncomingBounds;        // number of incoming bounds
    int         numOutgoingStates;        // total number of incoming states
    int         numOutgoingBounds;        // number of incoming bounds

public:
    /** constructor */
    CallbackStats(void) :
        numIncomingStates(0),
        numIncomingBounds(0),
        numOutgoingStates(0),
        numOutgoingBounds(0)
    { }

    /** destructor */
    ~CallbackStats() { }

    // member access methods
    int  getNumIncomingStates      (void) const { return numIncomingStates; }
    int  getNumIncomingBounds      (void) const { return numIncomingBounds; }
    int  getNumOutgoingStates      (void) const { return numOutgoingStates; }
    int  getNumOutgoingBounds      (void) const { return numOutgoingBounds; }

    // other methods: update statistics
    void incrementNumIncomingStates(const int n=1) { numIncomingStates += n; }
    void incrementNumIncomingBounds(const int n=1) { numIncomingBounds += n; }
    void incrementNumOutgoingStates(const int n=1) { numOutgoingStates += n; }
    void incrementNumOutgoingBounds(const int n=1) { numOutgoingBounds += n; }

    // other methods: print statistics
    void printStats(const DDX10_Base * const dd) const;
};


/** Native c++ code for the x10 class 'Solver'
 *
 *  Performs two types of functions:
 *   1. Implements pure virtual callback methods for its base class SolverX10Callback
 *      (unless they are solver specific, in which case solver specific classes
 *       inheriting from this class implement it)
 *   2. Implements routines that DDX10.x10 can use to interact with each solver
 *      (which, in most cases, call the appropriate routing in the solver object)
 */
class DDX10__DD : public x10::lang::X10Class, public DDX10_Callback {
protected:
    DDX10_Base                * dd;
    int                         solveStatus;
    std::string                 instanceName;

    CallbackStats               cb_stats;

    x10::lang::VoidFun_0_1<x10::array::Array<int>*>  * closure1;
    x10::lang::VoidFun_0_1<int>                      * closure2;
    x10::lang::VoidFun_0_1<int>                      * closure3;

public:
    RTT_H_DECLS_CLASS

    // X10 serialization requires these functions be stubbed out.
    // We never intend to serialize instances of DDX10_DD or
    // its subclasses, so stub them out here to raise errors if they
    // are called.
    virtual x10aux::serialization_id_t _get_serialization_id() { abort(); return -1;}
    virtual void _serialize_body(x10aux::serialization_buffer &) { abort(); }
        
    /** constructor
     *  @param _instanceName     name of the instance file
     */
    DDX10__DD(const char * const _instanceName) :
        dd(NULL),
        solveStatus(0),
        instanceName(_instanceName)
        { }

    /* destructor */
    virtual ~DDX10__DD() { if (dd != NULL) delete dd; }

    /* type 1: implement for base class DDX10_Callback */
    void x10_step(void) const {
        x10::lang::Runtime::probe();
    }

    /* type 1: implement for base class DDX10_Callback:
     *   process outgoing states by calling the appropriate closure to share them
     *
     *   @param states      a list of states to send
     */
    void x10_sendStates(const vector< vector<int> > & states);

    /* type 1: implement for base class DDX10_Callback:
     *   process outgoing lower bound by calling the appropriate closure to share it
     *
     *   @param states      a lower bound to send
     */
    void x10_sendLB(const int lb);

    /* type 1: implement for base class DDX10_Callback:
     *   process outgoing upper bound by calling the appropriate closure to share it
     *
     *   @param states      an upper bound to send
     */
    void x10_sendUB(const int ub);

    /** type 2: a method to set callbacks, to be called by DDX10.x10 */
    void attachCallbacks(x10::lang::VoidFun_0_1<x10::array::Array<int>*> * closure1_,
                         x10::lang::VoidFun_0_1<int>                     * closure2_,
                         x10::lang::VoidFun_0_1<int>                     * closure3_) {
        closure1 = closure1_;
        closure2 = closure2_;
        closure3 = closure3_;
    }

    /** type 2: main DD construction method, to be called by DDX10.x10 */
    void constructDD(x10::array::Array<int>* states_);

    /** type 2: when updated lower bound arrives from another place, process it */
    void updateLB(int lb_);

    /** type 2: when updated upper bound arrives from another place, process it */
    void updateUB(int ub_);

    /** type 2: kill method, to be called by DDX10.x10 */
    void kill(void) const { dd->x10_kill(); }

    /** type 2: print instance information */
    void printInstanceInfo(void) const;

    /** type 2: print final result (solver statistics, sat/unsat, plus solution if requested) */
    void printResults(const bool flag_printsoln = false) const;

    /** type 2: print solver statistics */
    void printStats(void) const { cb_stats.printStats(dd); dd->x10_printStats(); }

    /** type 2: print solution, assuming one has been found */
    void printSolution(void) const { assert(solver != NULL); dd->x10_printSolution(); }
};


///////////////////////////////
// INLINE METHODS
///////////////////////////////

// type 1: implement for base class DDX10_Callback:
//   process outgoing lower bound by calling the appropriate closure to share it
inline void DDX10__DD::x10_sendLB(const int lb) {
    // update x10 callback statistics
    cb_stats.incrementNumOutgoingBounds();

    // apply the appropriate closure
    // IMPORTANT NOTE: when passing control to X10 to execute the closure, X10 may
    // very well process incoming messages, including the kill() message and the
    // updateBounds() message! Hence, best to call the closure at the end
    // here, when we are not actively handling incoming or outgoing buffers
    x10::lang::VoidFun_0_1<int>::__apply(closure2, lb);
}

// type 1: implement for base class DDX10_Callback:
//   process outgoing lower bound by calling the appropriate closure to share it
inline void DDX10__DD::x10_sendUB(const int ub) {
    // update x10 callback statistics
    cb_stats.incrementNumOutgoingBounds();

    // apply the appropriate closure
    // IMPORTANT NOTE: when passing control to X10 to execute the closure, X10 may
    // very well process incoming messages, including the kill() message and the
    // updateBounds() message! Hence, best to call the closure at the end
    // here, when we are not actively handling incoming or outgoing buffers
    x10::lang::VoidFun_0_1<int>::__apply(closure2, ub);
}

// type 2: when updated bounds arrive from another place, process them
inline void DDX10__DD::updateLB(int lb) {
    dd->x10_updateLB(lb);
    // update x10 callback statistics
    cb_stats.incrementNumIncomingBounds();
}

// type 2: when updated bounds arrive from another place, process them
inline void DDX10__DD::updateUB(int ub) {
    dd->x10_updateUB(ub);
    // update x10 callback statistics
    cb_stats.incrementNumIncomingBounds();
}


#endif
