// -----------------------------------------------------------------
// Licensed Materials - Property of IBM
//
// DDX10
//
// (C) Copyright IBM Corporation 2013.  All Rights Reserved.
//
// Authors: Andre Cire, Ashish Sabharwal, Horst Samulowitz,
//          Vijay Saraswat
// -----------------------------------------------------------------


package com.ibm.ddx10;

import x10.io.Console;
import x10.lang.Math;
import x10.util.ArrayList;
import x10.util.ListIterator;
import x10.util.Pair;
import x10.util.Random;


public class StatePool {

	/** An internal class to store states: ub, lb, and a vector */
	protected static class State {
		protected var lb:Int = MINBOUND;
		protected var ub:Int = MAXBOUND;
		protected val vec    = new ArrayList[Int]();
	}

	/** An internal abstract class for storing the pool of states */
	protected interface IPool {
		public def isEmpty():Boolean;
		public def clear():void;
		public def size():Int;
		public def getLB(quick:Boolean):Int;   /** Lower bound, if known, across all states currently in the pool */
		public def getUB(quick:Boolean):Int;   /** Upper bound, if known, across all states currently in the pool */
		public def getAllStates():ArrayList[State];   /** REMOVE all states and return them in an ArrayList */
		public def add(State):void;
		public def get():State;
	}
	
	/** A generic class that PARTIALLY implements IPool; other pool classes
	 *  below extend this class and modify/fill-in the remaining functionality */
	protected static class GenericPool {
		protected var states:ArrayList[State];
		public def this(_initStates:ArrayList[State]) {
			states = (_initStates == null) ? new ArrayList[State]() : _initStates;
		}
		public def this() {	this(null);	}
		
		public def isEmpty():Boolean      { return states.isEmpty(); }
		public def clear():void           { states.clear(); }
		public def size():Int             { return states.size(); }
		public def getLB(quick:Boolean):Int {
			if (quick || isEmpty())
				return MINBOUND;
			// traverse the ArrayList and compute the lowest lower bound
			var _lb:Int = MAXBOUND;
			for (var i:Int=0; i<states.size(); i++)
				if (states(i).lb < _lb)
					_lb = states(i).lb;
			return _lb;
		}
		public def getUB(quick:Boolean):Int {
			if (quick || isEmpty())
				return MAXBOUND;
			// traverse the ArrayList and compute the highest upper bound
			var _ub:Int = MINBOUND;
			for (var i:Int=0; i<states.size(); i++)
				if (states(i).ub > _ub)
					_ub = states(i).ub;
			return _ub;
		}
		public def getAllStates():ArrayList[State] {
			val retValue = states;
			states = null;
			return retValue; 
		}
		public def add(s:State):void      { states.add(s); }		
	}
	
	/** An internal fully randomized implementation of state storage */
	protected static class RandomPool extends GenericPool implements IPool {
		protected val randNumGenerator = new Random();
		
		public def this(_initStates:ArrayList[State]) { super(_initStates); }
		public def this() {	this(null);	}
		
		public def get():State {
			val retStateID = randNumGenerator.nextInt(states.size());
			val retState = states(retStateID);
			if (retStateID == states.size()-1)
				states.removeLast();    // just remove the last element (it will be returned)
			else
				states(retStateID) = states.removeLast();  // move the last element to index retStateID 
			return retState;
		}
	}
	
	/** An internal stack (DFS) implementation of state storage */
	protected static class StackPool extends GenericPool implements IPool {
		public def this(_initStates:ArrayList[State]) { super(_initStates); }
		public def this() {	this(null);	}
		
		public def get():State { return states.removeLast(); }
	}
	
	/** An internal queue (BFS) implementation of state storage */
	protected static class QueuePool extends GenericPool implements IPool {
		public def this(_initStates:ArrayList[State]) { super(_initStates); }
		public def this() {	this(null);	}
		
		public def get():State { return states.removeFirst(); }
	}
	
	/** An internal implementation of state storage that combines BFS with
	 *  occassional scanning for the "best" states */
	protected static class OccasionallySortedQueuePool extends GenericPool implements IPool {
		protected var isMaximization:Boolean;
		protected var nCalls:Int = 0;
		protected val nSortingInterval = 10000;

		public def this(_isMaximization:Boolean, _initStates:ArrayList[State]) {
			super(_initStates);
			isMaximization = _isMaximization;
		}
		public def this(_isMaximization:Boolean) {
			this(_isMaximization, null);
		}

		public def get():State {
			if (nCalls % nSortingInterval == 0) {
				// sort the ArrayList by most promising upper bound first
				// Note: default sort is in increasing order
				if (isMaximization)
					states.sort((s1:State, s2:State) => s2.ub.compareTo(s1.ub));
				else
					states.sort((s1:State, s2:State) => s1.ub.compareTo(s2.ub));
			}
			return states.removeFirst();
		}
	}
	
	/** A priority queue (best-first) implementation of state storage */
	protected static class PQueuePool extends GenericPool implements IPool {
		// we will use use indices starting at 1
		// so that 2i and 2i+1 are the two children of i
		protected var isMaximization:Boolean;        /** is this a maximization problem or minimization */
		protected var randFrac:Double;               /** fraction of get() calls that return a random element */
		protected var randNumGenerator:Random;
		
		protected def isBetter(s1:State, s2:State):Boolean {
			// primary criterion  : more promising upper bound
			// secondary criterion: larger state vector size
			return (isMaximization && s1.ub > s2.ub)
			|| (!isMaximization && s1.ub < s2.ub)
			|| (s1.ub == s2.ub && s1.vec.size() > s2.vec.size());
		}
		protected def percolateUp(_i:Int) {
			var i:Int = _i;
			// if the parent is better, swap the elements, repeat
			while (i > 1 && isBetter(states(i), states(i/2))) {
				val tmp = states(i/2); states(i/2) = states(i); states(i) = tmp;   // swap
				i /= 2;   // move to the parent
			}			
		}
		protected def percolateDown(_i:Int) {
			var i:Int = _i;
			// if there is at least one child that is better, swap with it, repeat
			while (2*i < states.size()) {
				// find the best child
				var bestChild:Int = 2*i;
				if (2*i+1 < states.size() && isBetter(states(2*i+1), states(2*i)))
					bestChild = 2*i+1;
				if (!isBetter(states(bestChild), states(i)))
					break;
				val tmp = states(bestChild); states(bestChild) = states(i);	states(i) = tmp;   // swap
				i = bestChild;   // move to the child
			}
		}
		
		public def this(_isMaximization:Boolean, _randFrac:Double, _initStates:ArrayList[State]) {
			isMaximization = _isMaximization;
			randFrac       = _randFrac;
			if (randFrac > 0) {
				assert(randFrac >= 0 && randFrac <= 1);
				randNumGenerator = new Random();
			}
			states = new ArrayList[State]();
			// store a dummy state at index 0; we will use use indices starting at 1
			// so that 2i and 2i+1 are the two children of i
			states.add(new State());
			// now, if _states is non-null, add all elements of _states to the priority queue
			// NOTE: this takes O(N log N) and can be changed to a better, O(N), "makeHeap" operation
			if (_initStates != null) {
				for (var i:Int=0; i<_initStates.size(); i++) {
					states.add(_initStates(i));
					//if (randFrac < 1.0) percolateUp(states.size() - 1);
				}
			}
		}
		public def this(_isMaximization:Boolean, _randFrac:Double) {
			this(_isMaximization, _randFrac, null);
		}

		public def isEmpty():Boolean      { return states.size() == 1; }
		public def clear():void           { states.clear(); states.add(new State()); }
		public def size():Int             { return states.size() - 1; }

		public def getLB(quick:Boolean):Int {
			// for minimization problems, lower bound is on the top of the priority queue
			if (!isMaximization)
				return (isEmpty() ? MINBOUND : states(1).lb);
			// otherwise, resort to the full scan of states (starting at index 1)
			var _lb:Int = MAXBOUND;
			for (var i:Int=1; i<states.size(); i++)
				if (states(i).lb < _lb)
					_lb = states(i).lb;
			return _lb;
		}
		public def getUB(quick:Boolean):Int {
			// for maximization problems, upper bound is on the top of the priority queue
			if (isMaximization)
				return (isEmpty() ? MAXBOUND : states(1).ub);
			// otherwise, resort to the full scan of states (starting at index 1)
			var _ub:Int = MINBOUND;
			for (var i:Int=1; i<states.size(); i++)
				if (states(i).ub > _ub)
					_ub = states(i).ub;
			return _ub;
		}
		public def getAllStates():ArrayList[State] {
			val retValue = states;
			states = null;
			retValue.removeFirst();    // remove the dummy value at the front of the PQueue
			return retValue; 
		}
		public def add(s:State):void {
			// add the state to the end and percolate it up
			states.add(s);
			if (randFrac < 1.0) percolateUp(states.size() - 1);
		}
		public def get():State {
			if (states.size() == 2)   // got only one element
				return states.removeLast();
			
			// store the state at the top (or a random one, depending on randFrac) to return later
			var retStateID:Int = 1;
			// unclear whether nextDouble() returns a value in [0,1) or [0,1]; play safe
			if (randFrac >= 1 || (randFrac > 0 && randNumGenerator.nextDouble() < randFrac))
				retStateID = 1 + randNumGenerator.nextInt(this.size());
			val retState = states(retStateID);
			// remove the last element, move it to retStateID, and Percolate Down
			states(retStateID) = states.removeLast();
			if (randFrac < 1.0) percolateDown(retStateID);
			return retState;
		}
	}
	
	/** A priority queue + DFS implementation of state storage */
	protected static class PQueueDFSPool extends PQueuePool implements IPool {
		protected var poolLength:ULong;
		protected var dfsStates:ArrayList[State];
		protected var switchedToDFS:Boolean = false;
		
		public def this(_isMaximization:Boolean, _randFrac:Double, _initStates:ArrayList[State]) {
			super(_isMaximization, _randFrac, _initStates);
			dfsStates = new ArrayList[State]();
		}
		public def this(_isMaximization:Boolean, _randFrac:Double) {
			this(_isMaximization, _randFrac, null);
		}

		public def isEmpty():Boolean      { return (states.size() == 1 && dfsStates.size() == 0); }
		public def clear():void           { dfsStates.clear(); states.clear(); states.add(new State()); }
		public def size():Int             { return (states.size() + dfsStates.size() - 1); }
		
		public def getLB(quick:Boolean):Int {
			var _lb:Int = super.getLB(quick);
			if (!quick && !dfsStates.isEmpty()) {
				// if priority queue is empty, reset _lb to MAXBOUND, then loop over DFS states
				if (states.size() == 1)
					_lb = MAXBOUND;
				for (var i:Int=0; i<dfsStates.size(); i++)
					if (dfsStates(i).lb < _lb)
						_lb = dfsStates(i).lb;
			}
			return _lb;
		}
		public def getUB(quick:Boolean):Int {
			var _ub:Int = super.getUB(quick);
			if (!quick && !dfsStates.isEmpty()) {
				// if priority queue is empty, reset _ub to MINBOUND, then loop over DFS states
				if (states.size() == 1)
					_ub = MINBOUND;
				for (var i:Int=0; i<dfsStates.size(); i++)
					if (dfsStates(i).ub > _ub)
						_ub = dfsStates(i).ub;
			}
			return _ub;
		}
		public def getAllStates():ArrayList[State] {
			val retValue = states;
			states = null;
			retValue.removeFirst();    // remove the dummy value at the front of the PQueue
			retValue.addAll(dfsStates);   // add all states on the DFS stack
			dfsStates = null;
			return retValue; 
		}
		public def add(s:State):void {
			// check if priority queue is full
			if (poolLength >= MAXPQUEUELENGTH) {
				dfsStates.add(s);
				if (!switchedToDFS) {
					Console.OUT.println("PQueueDFSPool state pool SWITCHING to DFS; poolLength = " + poolLength);
					switchedToDFS = true;
				}
			} else {		
				// add the state to the end and percolate it up
				super.add(s);
				poolLength += s.vec.size();
			}
		}
		public def get():State {
			// give priority to nodes in the dfs list
			if (!dfsStates.isEmpty()) {
				return dfsStates.removeLast();
			}
			// if there is nothing there, revert to the priority queue as in the super class
			return super.get();
		}
	}
	
	
	protected static val MINBOUND         = -1000000;    // TODO: make this generic!!
	protected static val MAXBOUND         = +1000000;
	
	protected static val MAXPQUEUELENGTH  = 5500000000UL;
	
	protected var poolType:String;          /** the type of the pool: stack, queue, etc */
	protected var isMaximization:Boolean;   /** is this a minimization problem or maximization */
	protected var pool:IPool;               /** the actual pool of states */
	protected var maxPoolSizeReached:Int=0; /** maximum pool size reached over time */

	protected var lb:Int                  = MINBOUND;
	protected var ub:Int                  = MAXBOUND;
	protected var outstandingTasks:Int    = 0;
	protected var nStatesProcessed:Int    = 0;
	protected var nStatesPruned:Int       = 0;
	protected var nStatesAverageAdded:Int = 100;  // start with a value like 100
	
	// access methods
	public def getLB():Int                  { return lb; }
	public def getUB():Int                  { return ub; }
	public def getNOutstandingTasks():Int   { return outstandingTasks; }
	public def getPoolSize():Int            { return (pool != null ? pool.size() : 0); };
	public def getNStatesProcessed():Int    { return nStatesProcessed; };
	public def getNStatesPruned():Int       { return nStatesPruned; };
	public def getNStatesAverageAdded():Int { return nStatesAverageAdded; };
	public def getMaxPoolSizeReached():Int  { return maxPoolSizeReached; };
	public def getPoolLB(quick:Boolean):Int {
		return ((pool != null && !pool.isEmpty()) ? pool.getLB(quick) : lb);
	}
	public def getPoolUB(quick:Boolean):Int {
		// ASHISH: commenting out "updateUB(_ub); it should NOT be called here, I think,
		// because of Outstanding Tasks which may still create child nodes with higher UB than this
		// updateUB(_ub);
		return ((pool != null && !pool.isEmpty()) ? pool.getUB(quick) : ub);
	}
	
		
	// general methods
	public def updateLB(_lb:Int):void       { if (_lb > lb) lb = _lb; }
	public def updateUB(_ub:Int):void       { if (_ub < ub) ub = _ub; }
	public def decrOutstandingTasks(_n:Int):void { outstandingTasks -= _n; }
	
	
	/** Clear the state pool */
	public def clear() {
		if (pool != null) {
			poolType = "";
			pool.clear();
		}
	}
	
	/** Creates a new state pool and starts it off with the default initial
	 *  state as the only state (unless the set of initial states is provided) */
	public def createPool(_poolType:String, _isMaximization:Boolean, _randFrac:Double, _initStates:ArrayList[State]) {
		poolType = _poolType;
		isMaximization = _isMaximization;
		assert(_randFrac >= 0.0 && _randFrac <= 1.0);

		if (_poolType.equals(""))
			pool = null;
		else if (_poolType.equals("RandomPool"))
			pool = new RandomPool(_initStates);
		else if (_poolType.equals("StackPool"))
			pool = new StackPool(_initStates);
		else if (_poolType.equals("QueuePool"))
			pool = new QueuePool(_initStates);
		else if (_poolType.equals("OccasionallySortedQueuePool"))
			pool = new OccasionallySortedQueuePool(isMaximization, _initStates);
		else if (_poolType.equals("PQueuePool"))
			pool = new PQueuePool(isMaximization, _randFrac, _initStates);
		else if (_poolType.equals("PQueueDFSPool"))
			pool = new PQueueDFSPool(isMaximization, _randFrac, _initStates);
		else
			throw new Exception("ERROR: Unknown poolType " + _poolType + " in StatePool constructor. Aborting.");

		// If no initial states are provided, add a default initial state to the pool
		if (_initStates == null) {
			pool.add(new State());
			if (maxPoolSizeReached == 0)
				maxPoolSizeReached = 1;
			}
	}
	public def createPool(_poolType:String, _isMaximization:Boolean, _randFrac:Double) {
		createPool(_poolType, _isMaximization, _randFrac, null);
	}
	public def createPool(_poolType:String, _isMaximization:Boolean, _states:ArrayList[State]) {
		createPool(_poolType, _isMaximization, 0.0, _states);
	}
	public def createPool(_poolType:String, _isMaximization:Boolean) {
		createPool(_poolType, _isMaximization, 0.0, null);
	}

	/** Change the type of the pool for future processing, while keeping existing states */
	public def changePoolType(_poolType:String, _randFrac:Double) {
		assert(_poolType != poolType);
		val currentStates = (pool == null) ? null : pool.getAllStates();
		createPool(_poolType, isMaximization, _randFrac, currentStates);
	}
	
	/** Get a list of states from the state pool and also increment
	 *  the number of outstanding tasks; ATOMIC METHOD
	 * 
	 *  @param n           max number of states to return
	 *  @return            first: -1 if no more states and no outstanding tasks
	 *                     otherwise number of states returned; second: states
	 */
	public atomic def getStates(_n:Int):Pair[Int,Rail[Int]] {
		if (pool.isEmpty() && outstandingTasks == 0)
			return new Pair[Int,Rail[Int]](-1, new Rail[Int](0));
		
		var nStatesToSend:Int = _n;
		if (_n == -1) {
			// nStatesToSend = Int.operator_as(Math.min(nStatesAverageAdded / 2.0, (pool.size()*1.0 / Place.MAX_PLACES)*2.0));
			nStatesToSend = Int.operator_as(Math.min(nStatesAverageAdded, (pool.size()*1.0 / Place.MAX_PLACES)*2.0));
			nStatesToSend = Math.max(10, nStatesToSend);
		}
		//Console.OUT.println("States to send = " + nStatesToSend);
		
		// select up to n states to return; also increment outstandingTasks
		// TODO: not sure how to have dynamic Rail[Int] size in X10, so first
		//       creating a tmp ArrayList object and then converting it to Rail[Int]
		var tmpStates:ArrayList[Int] = new ArrayList[Int]();
		var i:Int = 0;
		while (!pool.isEmpty() && i<nStatesToSend) {
			val state = pool.get();
			//Console.OUT.println("State UB = " + state.ub);
			if ((isMaximization && state.ub <= lb) || (!isMaximization && state.lb <= ub))  {
				// discard this state: a better feasible solution is already known!
				nStatesPruned++;
				continue;
			}
			tmpStates.add(2 + state.vec.size());
			tmpStates.add(state.ub);
			tmpStates.add(state.lb);
			for (var j:Int=0; j<state.vec.size(); j++)
				tmpStates.add(state.vec(j));
			++i;
		}
		
		//ListIterator<ArrayList> iterArrayList = pool.iterator();
		//while(iterArrayList.hasNext()){
		//		val state = iterArrayList.next();
		//	}
		
		val states = new Rail[Int](tmpStates.size());
		for (var j:Int=0; j<tmpStates.size(); j++)
			states(j) = tmpStates(j);
		
		outstandingTasks += i;
		
		// Return selected state(s)
		return new Pair[Int,Rail[Int]](i, states);
	}
	
	/** Add given states to the state pool */
	public atomic def addStates(_states:Rail[Int]){
		var i:Int=0;
		var nStatesAdded:Int=0;
		while (i<_states.size) {
			val stateVecSize = _states(i++) - 2;
			assert(stateVecSize >= 0);
			var state:State = new State();
			state.ub = _states(i++);
			state.lb = _states(i++);
			if ((isMaximization && state.ub <= lb) || (!isMaximization && state.lb <= ub)) {
				// discard this state: a better feasible solution is already known!
				i += stateVecSize;
				nStatesPruned++;
			}
			else {
				for (var j:Int=0; j<stateVecSize; j++)
					state.vec.add(_states(i++));
				pool.add(state);
				nStatesAdded++;
				//Console.OUT.println(state.toString());
			}
		}
		val r = 0.5;
		nStatesAverageAdded = Int.operator_as(r * nStatesAverageAdded + (1-r) * nStatesAdded);
		nStatesProcessed += nStatesAdded;
		if (pool.size() > maxPoolSizeReached)
			maxPoolSizeReached = pool.size();
	}
}
